/*
 * Copyright (C) 2022 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "../includes/common.h"
#include <nfc_api.h>
#include <nfc_int.h>
#include <rw_int.h>
#include <stdlib.h>
#include <string.h>
#include <tags_defs.h>

#define T3T_MSG_FELICALITE_MC_OFFSET 0x01

bool testInProgress = false;

struct sigaction new_action, old_action;

void sigabrt_handler(int signum, siginfo_t *info, void *context) {
  if (testInProgress && info->si_signo == SIGABRT) {
    (*old_action.sa_sigaction)(signum, info, context);
    return;
  }
  exit(EXIT_FAILURE);
}

extern tRW_CB rw_cb;
extern tNFC_CB nfc_cb;
tNFC_CONN *p_data;
void rw_init(void);
tNFC_STATUS rw_t3t_select(uint8_t peer_nfcid2[NCI_RF_F_UID_LEN],
                          uint8_t mrti_check, uint8_t mrti_update);

void *allocate_memory(size_t size) {
  void *ptr = malloc(size);
  memset(ptr, 0x0, size);
  return ptr;
}

/* States */
enum {
  RW_T3T_STATE_NOT_ACTIVATED,
  RW_T3T_STATE_IDLE,
  RW_T3T_STATE_COMMAND_PENDING
};

/* Enumeration of API commands */
enum {
  RW_T3T_CMD_DETECT_NDEF,
  RW_T3T_CMD_CHECK_NDEF,
  RW_T3T_CMD_UPDATE_NDEF,
  RW_T3T_CMD_CHECK,
  RW_T3T_CMD_UPDATE,
  RW_T3T_CMD_SEND_RAW_FRAME,
  RW_T3T_CMD_GET_SYSTEM_CODES,
  RW_T3T_CMD_FORMAT,
  RW_T3T_CMD_SET_READ_ONLY_SOFT,
  RW_T3T_CMD_SET_READ_ONLY_HARD,
  RW_T3T_CMD_MAX
};

/* Sub-states */
enum {
  /* Sub states for formatting Felica-Lite */
  RW_T3T_FMT_SST_POLL_FELICA_LITE, /* Waiting for POLL Felica-Lite response (for
   formatting) */
  RW_T3T_FMT_SST_CHECK_MC_BLK,     /* Waiting for Felica-Lite MC (MemoryControl)
       block-read to complete */
  RW_T3T_FMT_SST_UPDATE_MC_BLK,    /* Waiting for Felica-Lite MC (MemoryControl)
      block-write to complete */
  RW_T3T_FMT_SST_UPDATE_NDEF_ATTRIB, /* Waiting for NDEF attribute block-write
   to complete */

  /* Sub states for setting Felica-Lite read only */
  RW_T3T_SRO_SST_POLL_FELICA_LITE, /* Waiting for POLL Felica-Lite response (for
   setting read only) */
  RW_T3T_SRO_SST_UPDATE_NDEF_ATTRIB, /* Waiting for NDEF attribute block-write
   to complete */
  RW_T3T_SRO_SST_CHECK_MC_BLK, /* Waiting for Felica-Lite MC (MemoryControl)
   block-read to complete */
  RW_T3T_SRO_SST_UPDATE_MC_BLK /* Waiting for Felica-Lite MC (MemoryControl)
   block-write to complete */
};

void poc_cback(tRW_EVENT event, tRW_DATA *p_rw_data) {
  (void)event;
  (void)p_rw_data;
}

void GKI_start_timer(uint8_t, int32_t, bool) {}

void GKI_stop_timer(uint8_t) {}

void GKI_freebuf(void *) {}

void exit_handler(void) {
  if (p_data) {
    if (p_data->data.p_data) {
      free(p_data->data.p_data);
      p_data->data.p_data = nullptr;
    }
    free(p_data);
    p_data = nullptr;
  }
}

int main() {
  atexit(exit_handler);
  sigemptyset(&new_action.sa_mask);
  new_action.sa_flags = SA_SIGINFO;
  new_action.sa_sigaction = sigabrt_handler;
  sigaction(SIGABRT, &new_action, &old_action);

  tNFC_ACTIVATE_DEVT p_activate_params = {};
  p_activate_params.protocol = NFC_PROTOCOL_ISO_DEP;
  p_activate_params.rf_tech_param.mode = NFC_DISCOVERY_TYPE_POLL_A;
  RW_SetActivatedTagType(&p_activate_params, &poc_cback);
  FAIL_CHECK(rw_cb.p_cback == &poc_cback);

  tRW_T3T_CB *p_t3t = &rw_cb.tcb.t3t;

  GKI_init();
  rw_init();
  rw_cb.p_cback = &poc_cback;

  uint8_t peer_nfcid2[NCI_RF_F_UID_LEN];
  uint8_t mrti_check = 1, mrti_update = 1;
  FAIL_CHECK(rw_t3t_select(peer_nfcid2, mrti_check, mrti_update) ==
             NFC_STATUS_OK)

  p_data = (tNFC_CONN *)allocate_memory(sizeof(tNFC_CONN));
  FAIL_CHECK(p_data);

  p_data->data.p_data = (NFC_HDR *)allocate_memory(sizeof(NFC_HDR) * 3);
  FAIL_CHECK(p_data->data.p_data);

  p_data->status = NFC_STATUS_OK;

  p_t3t->cur_cmd = RW_T3T_CMD_CHECK_NDEF;
  p_t3t->rw_state = RW_T3T_STATE_COMMAND_PENDING;
  p_t3t->flags |= RW_T3T_FL_IS_FINAL_NDEF_SEGMENT;
  p_t3t->ndef_attrib.ln = 0x000F;

  NFC_HDR *p_msg = (p_data->data).p_data;
  p_msg->offset = 0;
  p_msg->len = T3T_MSG_RSP_OFFSET_CHECK_DATA + 1;

  uint8_t *p_t3t_rsp = (uint8_t *)(p_msg + 1) + p_msg->offset;
  p_t3t_rsp[0] = NCI_STATUS_OK;
  p_t3t_rsp++;
  p_t3t_rsp[T3T_MSG_RSP_OFFSET_RSPCODE] = T3T_MSG_OPC_CHECK_RSP;
  p_t3t_rsp[T3T_MSG_RSP_OFFSET_STATUS1] = T3T_MSG_RSP_STATUS_OK;
  p_t3t_rsp[T3T_MSG_RSP_OFFSET_NUMBLOCKS] = 0;

  tNFC_CONN_CB *p_cb = &nfc_cb.conn_cb[NFC_RF_CONN_ID];
  tNFC_CONN_EVT event = NFC_DATA_CEVT;
  memcpy(p_t3t->peer_nfcid2, &p_t3t_rsp[T3T_MSG_RSP_OFFSET_IDM],
         NCI_NFCID2_LEN);
  testInProgress = true;
  p_cb->p_cback(0, event, p_data);
  testInProgress = false;

  return EXIT_SUCCESS;
}
